package fx.animation;

import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.util.LinkedList;
import java.util.List;

import control.BridgeControl;
import fx.options.OptionsController;
import fx.options.OptionsSet;
import javafx.animation.Animation;
import javafx.animation.KeyFrame;
import javafx.animation.Timeline;
import javafx.event.ActionEvent;
import javafx.event.EventHandler;
import javafx.scene.Group;
import javafx.scene.control.Alert;
import javafx.scene.control.Alert.AlertType;
import javafx.scene.layout.Pane;
import javafx.util.Duration;
import vehicle.VehicleImpl;
import vehicle.Vehicle.VOrigin;

public class AnimationController {

	//Only set/created once
	private AnimationView view;
	private OptionsController optionsController;
	
	//BridgeControl for the current run
	private BridgeControl bridgeControl;
	
	//settings which could be changed using the Options Tab
	private double maxLoad;
	private double spawnRate;
	private boolean useWeightOne;
	private boolean lowrider;


	//All vehicles are stored in a List (in order to access the methods of the Vehicles)
	//AND all vehicles are assigned to vehicleShapes Group (in order to show them)
	private Group vehicleShapes;
	private List<VehicleImpl> vehicleList;

	//used by the refresh Method and the scores
	private double load;
	private double allocatedLoad;
	private int corsaWest;
	private int corsaEast;
	private boolean corsaDuel;
	private boolean vehicleWaiting;
	private long frame;
	
	//only used by the scores
	private double efficiencyScore;
	private double fairnessScore;
	private double[] waitingTimeLast10Vehicles  = new double[10];
	private int passedVehicles;

	//used to refresh, manage and terminate each run.
	private Timeline currentTimeline;
	private boolean running;	//true if the timeline is active
	private boolean collapsed;  //true if the bridge collapsed
	private int nextFreeVehicleID;

	/**
	 * Creates the AnimationView, stores the optionsController
	 * and assigns a function to the Button
	 */
	public AnimationController(OptionsController optionsController) {
		this.optionsController = optionsController;
		this.view = new AnimationView();
		this.vehicleShapes = view.getVehicleShapes();

		view.getStartBTN().setOnAction(new StartBTNHandler());
	}

	
	/**
	 * Triggered whenever the ReStart Button is used
	 */
	class StartBTNHandler implements EventHandler<ActionEvent> {
		public void handle(ActionEvent e) {
			startNewRun();
		}
	}


	/**
	 * Called whenever the ReStart Button is used. 
	 * Loads settings, terminates the current run, resets all values, starts a new run.
	 */
	private void startNewRun() {
		OptionsSet optionsSet = optionsController.getSettings();
		try {
			
			//set options and initialize the BridgeControl
			this.bridgeControl = initBridgeControl(optionsSet.getMutexClass());

			this.maxLoad = optionsSet.getMaxLoad();
			this.spawnRate = optionsSet.getSpawnRate() / 1000.0;
			this.useWeightOne = optionsSet.getUseWeightOne();
			this.lowrider = optionsSet.getLowrider();
			
			this.bridgeControl.init(maxLoad);
			
			//stop current run (if existing) and cleans the view
			if (running) {
				currentTimeline.stop();
				view.resetAnimation();
				closeInfoDialogs();
			}

			//creates a new vehicle list for the next run
			vehicleList = new LinkedList<VehicleImpl>();

			//reset all values
			frame = 0;
			efficiencyScore = 0;
			fairnessScore = 0;
			collapsed = false;
			nextFreeVehicleID = 0;
			passedVehicles = 0;

			running = true;

			//starts new timeline. The timeline will trigger the refresh method periodically 
			KeyFrame keyFrame1 = new KeyFrame(Duration.millis(1000 / optionsSet.getFps()));
			currentTimeline = new Timeline(new KeyFrame(Duration.seconds(0), evt -> {
				refresh();
			}), keyFrame1);
			currentTimeline.setCycleCount(Animation.INDEFINITE);
			currentTimeline.play();

		} catch (ClassNotFoundException | NoSuchMethodException | NullPointerException | SecurityException
				| InstantiationException | IllegalAccessException | IllegalArgumentException
				| InvocationTargetException e) {

			// This exceptions are caused by the reflection (also a NullPointer is caught ... dirty)
			Alert alert = new Alert(AlertType.ERROR);
			alert.setTitle("Error");
			alert.setHeaderText("Reflection failed");
			alert.setContentText("Could not load \"" + optionsSet.getMutexClass() + "\".");
			alert.showAndWait();
		}
	}

	/**
	 * Initializes the BridgeControl by using the java reflection api.
	 * 1. Searching for a class with the denoted name
	 * 2. Get the default constructor
	 * 3. Using the constructor to instantiate the BridgeControl class
	 */
	private BridgeControl initBridgeControl(String classname)
			throws ClassNotFoundException, NoSuchMethodException, SecurityException, InstantiationException,
			IllegalAccessException, IllegalArgumentException, InvocationTargetException {
		Class<?> c = Class.forName(classname);
		Constructor<?> constructor = c.getConstructor();
		BridgeControl bridgeControl = (BridgeControl) constructor.newInstance();
		return bridgeControl;
	}

	/**
	 * Called by the timeline. 
	 * 1. Increment frame count
	 * 2. Refresh all vehicles look wether or not
	 * 		- the bridge is overloaded
	 * 		- there is at least one car waiting (important for E-Score)
	 * 		- there is a lowrider battle
	 * 		- how many weight is already "allocated" to the bridge
	 * 		  (sum the weight of all vehicles in state DRIVENOFF and ONBRIDGE)
	 * 		  (important for E-Score)
	 */
	public void refresh() {
		frame += 1;
		LinkedList<VehicleImpl> toRemove = new LinkedList<VehicleImpl>();
		for (VehicleImpl v : vehicleList) {
			switch (v.refresh()) {
			case WAITING:
				vehicleWaiting = true;
				break;
			case DRIVENOFF:
				allocatedLoad += v.getWeight();
				break;

			case ONBRIDGE:
				load += v.getWeight();
				allocatedLoad += v.getWeight();
				if (v.isLowrider()) {
					if (v.getOrigin() == VOrigin.EAST) {
						corsaEast++;
					} else {
						corsaWest++;
					}
					if (corsaDuel)
						v.setAssiMode();
				}
				break;
			case DRIVENAWAY:
				waitingTimeLast10Vehicles[passedVehicles++ % 10] = v.getTimeWaited();
				toRemove.add(v);
				vehicleShapes.getChildren().remove(v);
				break;
			default:
				break;
			}
		}
		vehicleList.removeAll(toRemove);

		if (!collapsed) {
			if (Math.random() < spawnRate) {
				spawnVehicle();
			}
			updateScoreEfficiency();
			updateScoreFairness();

			refreshDisplays();
			if (load > maxLoad) {
				overloadHandler();
			}
			corsaDuel = (corsaWest * corsaEast != 0);
		}

		load = 0;
		allocatedLoad = 0;
		corsaWest = 0;
		corsaEast = 0;
		vehicleWaiting = false;

	}

	/**
	 * Refresh the frame counter and the two score displays.
	 */
	private void refreshDisplays() {
		view.setLoadTXTcontent(String.format("Load: %.02f t / %.02f t", load, maxLoad));
		view.setFramesTXTcontent(String.format("%d", frame));

		if (frame % 100 == 0 && frame != 0)
			view.setScoreTXTcontent(String.format("Efficiency: %.0f  Fairness: %.0f  ", efficiencyScore, fairnessScore));
		if (frame % 5000 == 0 && frame != 0)
			view.setScore2TXTcontent(String.format("Efficiency: %.0f  Fairness: %.0f  (Frame %d)", efficiencyScore, fairnessScore, frame));
	}

	/**
	 * Called by refreh() whenever the maximum weight was exceeded.
	 */
	private void overloadHandler() {
		view.setOverloadTXTcontent("OVERLOADED");
		view.hideBridgeRECT();
		for (VehicleImpl v : vehicleList)
			v.informAboutCollapse();
		collapsed = true;
	}

	/**
	 * Called by refresh() in each iteration. Randomly creates new Vehicle.
	 */
	private void spawnVehicle() {
		VOrigin org = (Math.random() < 0.5) ? VOrigin.EAST : VOrigin.WEST;
		double weight = useWeightOne ? 1 : 0.8 + Math.random();
		boolean corsaB = lowrider && (Math.random() < 0.1);

		VehicleImpl tmp = new VehicleImpl(org, weight, bridgeControl, corsaB, nextFreeVehicleID++);

		vehicleShapes.getChildren().add(tmp);
		vehicleList.add(tmp);
	}

	/**
	 * Refresh the efficiency score. Called by the refresh Method.
	 */
	private void updateScoreEfficiency() {
		if(corsaDuel){
			efficiencyScore -= 1;
		} else if (vehicleWaiting) {
			efficiencyScore += (allocatedLoad / maxLoad);
		} else {
			efficiencyScore += 1;
		} 
	}

	/**
	 * Refresh the fairness score. Called by the refresh Method.
	 */
	private void updateScoreFairness() {
		if (passedVehicles > 10 && frame % 100 == 0) {
			double sum = 0;
			double max = 0;
			for (double l : waitingTimeLast10Vehicles) {
				sum += l;
				max = Math.max(l, max);
			}
			sum /= 10;
			fairnessScore += 100 * (sum/max);
		}
	}
	
	/**
	 * Called whenever the PrimaryStage is closed. Redirects the call
	 * to all existing vehicles.
	 */
	public void closeInfoDialogs() {
		if (running) {
			for (VehicleImpl v : vehicleList)
				v.closeInfoDialog();
		}
	}

	/**
	 * Used once by BridgeFX in order to get the AnimationView
	 */
	public Pane getView() {
		return view.getRoot();
	}

}
