package fx.options;

import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.text.NumberFormat;
import java.util.Properties;

import javafx.fxml.FXML;
import javafx.scene.control.Button;
import javafx.scene.control.CheckBox;
import javafx.scene.control.ComboBox;
import javafx.scene.control.Slider;
import javafx.scene.control.TextField;
import javafx.scene.layout.Pane;

/**
 * @author Thierry Meurers
 *
 * Controller class for the option tab. All fields are injected by fxml and are tied 
 * to the layout defined in the BridgeOptions.fxml file.
 */
public class OptionsController {

	@FXML
	private Slider fpsSL;
	@FXML
	private TextField spawnRateTF;
	@FXML
	private CheckBox lowriderCB;
	@FXML
	private TextField maxLoadTF;
	@FXML
	private CheckBox weightOneCB;
	@FXML
	private Slider maxLoadSL;
	@FXML
	private Button defaultsBTN;
	@FXML
	private TextField fpsTF;
	@FXML
	private Button applyBTN;
	@FXML
	private Slider spawnRateSL;
	@FXML
	private ComboBox<String> mutexClassCB;
	@FXML
	private Pane optionsPane;

	
	/**
	 * Binds the value of the 3 textfields to the 3 sliders
	 */
	public void initialize() {
		fpsTF.textProperty().bindBidirectional(fpsSL.valueProperty(), NumberFormat.getNumberInstance());
		maxLoadTF.textProperty().bindBidirectional(maxLoadSL.valueProperty(), NumberFormat.getNumberInstance());
		spawnRateTF.textProperty().bindBidirectional(spawnRateSL.valueProperty(), NumberFormat.getNumberInstance());
	}

	/**
	 * This method is called on start by the BridgeFX-class. All mutex classes are,
	 * defined in BridgeFX, are set to the Combobox.
	 * 
	 * @param mutexClasses	Array of possible mutex classes
	 */
	public void setMutexClasses(String[] mutexClasses){
		mutexClassCB.getItems().addAll(mutexClasses);
		if(mutexClasses.length > 0){
			mutexClassCB.setValue(mutexClasses[0]);
		}
	}
	
	/**
	 * Gathers all settings and returns a new OptionsSet.
	 * 
	 * @return	OptionsSet with all current values
	 */
	public OptionsSet getSettings() {
		return new OptionsSet(mutexClassCB.getValue(), fpsSL.getValue(), maxLoadSL.getValue(), spawnRateSL.getValue(),
				weightOneCB.isSelected(), lowriderCB.isSelected());
	}
	
	/**
	 * All settings (except for the mutex class) are saved to a external file.
	 * 
	 * Called whenever the PrimaryStage is closed.
	 */
	public void saveOptions() {
		try (OutputStream output = new FileOutputStream("BridgeFX.properties")) {

			Properties prop = new Properties();

			prop.setProperty("fps", fpsSL.getValue() + "");
			prop.setProperty("maxLoad", maxLoadSL.getValue() + "");
			prop.setProperty("spawnRate", spawnRateSL.getValue() + "");
			prop.setProperty("weightOne", weightOneCB.isSelected() + "");
			prop.setProperty("lowrider", lowriderCB.isSelected() + "");

			prop.store(output, "BridgeFX Options File");

		} catch (IOException e) {
			e.printStackTrace();
		}
	}
	


	
	/**
	 * Read all options from external file to the GUI.
	 */
	public void readOptions(){
		Properties prop = new Properties();
		
        try (InputStream input = new FileInputStream("BridgeFX.properties")) {
 
            prop.load(input);
 
            double fps = Double.valueOf(prop.getProperty("fps"));
            double maxLoad = Double.valueOf(prop.getProperty("maxLoad"));
            double spawnRate = Double.valueOf(prop.getProperty("spawnRate"));
            boolean weightOne = Boolean.valueOf(prop.getProperty("weightOne"));
            boolean lowrider = Boolean.valueOf(prop.getProperty("lowrider"));
            
            setAll(fps, maxLoad, spawnRate, weightOne, lowrider);
 
        } catch (IOException e) {
            System.err.println("Could not read (all) Options from file (BridgeFX.properties)");
        }
        
        
	}
	
	/**
	 * Restores the default settings.
	 */
	public void restoreDefaults() {
		setAll(75,10,20,false,false);
	}
	
	/**
	 * Sets the passed values to the GUI.
	 * Used by readOptions() and restoreDefaults().
	 */
	private void setAll(double fps, double maxLoad, double spawnRate, boolean weightOne, boolean lowrider){
		fpsSL.setValue(fps);
		maxLoadSL.setValue(maxLoad);
		spawnRateSL.setValue(spawnRate);
		weightOneCB.setSelected(weightOne);
		lowriderCB.setSelected(lowrider);
		
	}
}