package vehicle;

import java.util.Random;

import control.BridgeControl;
import javafx.application.Platform;
import javafx.beans.property.SimpleStringProperty;
import javafx.beans.property.StringProperty;
import javafx.event.EventHandler;
import javafx.geometry.Insets;
import javafx.scene.paint.Color;
import javafx.scene.shape.Rectangle;
import javafx.stage.Stage;
import javafx.stage.WindowEvent;
import javafx.scene.Scene;
import javafx.scene.control.Label;
import javafx.scene.control.TextArea;
import javafx.scene.input.MouseEvent;
import javafx.scene.layout.BorderPane;

/**
 * @author Thierry Meurers
 * 
 * Implementation of Vehicle. Extends Rectangle and is directly added to the
 * Animation view. A list of all vehicles is maintained by AnimationView.
 *
 */
public class VehicleImpl extends Rectangle implements Vehicle {
	
	
	//informations about the scenario
	final static private double startWest = -50;
	final static private double startEast = 1050;
	final static private double borderWest = 300;
	final static private double borderEast = 700;
	final static private double gatherPointDistance = 160;
	final static private double height = 300; 
	
	//dimensions of the vehicle and all possible colors.
	final static private Color[] clr = { Color.RED, Color.AQUA, Color.MAGENTA, Color.GREENYELLOW, Color.ORANGERED,
			Color.MEDIUMSLATEBLUE };
	final static private double vHeight = 16;
	final static private double vLength = 32;
	
	
	//defining all possible states ...
	public enum VStatus {
		ARRIVING, WAITING, DRIVENOFF, ONBRIDGE, PASSED, DRIVENAWAY, SHOCKED, FALLING
	}
	
	//informations about scenario which are (almost) unique for each vehicle
	private double waitingPoint;
	private double bridgeBegin;
	private double bridgeEnd;
	private double landscapeEnd;
	
	private BridgeControl bridgeControl;
	
	//static, unique properties of vehicle
	private Double weight;
	private VOrigin origin;
	private double velocityX;
	private double velocityY;
	private boolean fromWest;
	private boolean corsaB;

	//dynamic, unique properties  of vehicle
	private double xPos;
	private double yPos;
	private VStatus status;
	private double accelerationY;
	private int jumpCounter;
	private boolean paused = true;
	private long timeWaited = 0;
	
	//Informations used by the info dialog
	private Stage infoDialog;
	private StringProperty statusProperty = new SimpleStringProperty("ARRIVING");
	private StringProperty messageProperty = new SimpleStringProperty("<no message>");
	private long id;
	private boolean infoDialogOpened = false;



	/**
	 * @param origin
	 * @param weight
	 * @param bc
	 * @param bs
	 */
	public VehicleImpl(VOrigin origin, Double weight, BridgeControl bc, boolean corsaB, long id) {
		super();

		this.id = id;

		Random rnd = new Random();

		this.setHeight(vHeight);
		this.setWidth(vLength);
		this.setArcHeight(10);
		this.setArcWidth(10);
		this.setStroke(Color.BLACK);
		if (corsaB) {
			this.setFill(Color.DEEPPINK);
		} else {
			this.setFill(clr[rnd.nextInt(clr.length)]);
		}

		if (origin == VOrigin.EAST) {
			this.velocityX = -1.5;
			this.waitingPoint = borderEast + gatherPointDistance
					- rnd.nextDouble() * gatherPointDistance;
			this.bridgeBegin = borderEast - (vLength / 2);
			this.bridgeEnd = borderWest - (vLength / 2);
			this.landscapeEnd = startWest;
			this.xPos = startEast;
			this.fromWest = false;

		} else {
			this.velocityX = 1.5;
			this.waitingPoint = borderWest - gatherPointDistance
					+ rnd.nextDouble() * gatherPointDistance - vLength;
			this.bridgeBegin = borderWest - (vLength / 2);
			this.bridgeEnd = borderEast - (vLength / 2);
			this.landscapeEnd = startEast;
			this.xPos = startWest;
			this.fromWest = true;
		}

		this.velocityY = 1;
		this.yPos = height - vHeight;

		this.setY(yPos);
		this.setX(xPos);

		this.origin = origin;
		this.bridgeControl = bc;
		this.weight = weight;
		this.corsaB = corsaB;

		this.status = VStatus.ARRIVING;

		this.setOnMouseClicked(new EventHandler<MouseEvent>() {
			@Override
			public void handle(MouseEvent t) {
				if (!infoDialogOpened) {
					infoDialogOpened = true;
					infoDialog = createInfoDialog();
					infoDialog.show();
				}
			}
		});
	}

	//Getters and Setters visible for BridgeControl
	@Override
	public Double getWeight() {
		return weight;
	}

	@Override
	public VOrigin getOrigin() {
		return origin;
	}
	
	@Override
	public boolean isLowrider() {
		return corsaB;
	}

	@Override
	public long getVehicleId() {
		return id;
	}
	
	@Override
	public void setMessage(String msg) {
		Platform.runLater(new Runnable() {
		    public void run() {
		    	messageProperty.set(msg);
		    }
		});
	}
	
	
	//Getters, Setters, small Methods visible for AnimationControl
	public Double getXPos() {
		return xPos;
	}
	
	public double getTimeWaited() {
		return timeWaited;
	}

	/**
	 * Is called by AnimationController on lowrider-vehicles once a
	 * "Lowrider-Battle" is started. 
	 */
	public void setAssiMode() {
		if (jumpCounter == 20) {
			velocityX = fromWest ? 0.3 : -0.3;
			velocityY = 5;
		}
	}
	
	public void closeInfoDialog() {
		if(infoDialogOpened)
			infoDialog.close();
		infoDialogOpened = false;
	}

	/**
	 * Called by AnimationController once the Bridge collapsed.
	 * All vehicles on the Bridge will enter the state FALLING.
	 * All other veicles (except for the vehicles which has already
	 * passed the bridge) will be freezed by seting them to SHOCKED state
	 */
	public void informAboutCollapse() {
		switch (status) {
		case WAITING:
		case ARRIVING:
		case DRIVENOFF:
			status = VStatus.SHOCKED;
			statusProperty.set("SHOCKED");
			break;
		case ONBRIDGE:
			status = VStatus.FALLING;
			statusProperty.set("FALLING");
			break;
		default:
			break;
		}
	}

	/**
	 * This method is periodically called by the AnimatonController.
	 * In each step it checks the current state of the vehicle and
	 * calculates the new state, x-position and y-position.
	 * 
	 * @return  Current status of vehicle (at the end of method call)
	 */
	public VStatus refresh() {
		switch (status) {
		case ARRIVING:
			if (fromWest && xPos < waitingPoint || !fromWest && waitingPoint < xPos) {
				xPos += velocityX;
			} else {
				status = VStatus.WAITING;
				statusProperty.set("WAITING");
				Thread t1 = new Thread(new BridgeControlRequestor());
				t1.setDaemon(true);
				t1.start();
			}
			break;
		case WAITING:
			statusProperty.set("WAITING (" + timeWaited++ + ")");
			if (!paused) {
				status = VStatus.DRIVENOFF;
				statusProperty.set("DRIVENOFF (" + timeWaited + ")");
			}
			break;
		case DRIVENOFF:
			if (fromWest && xPos < bridgeBegin || !fromWest && bridgeBegin < xPos) {
				xPos += velocityX;
			} else {
				status = VStatus.ONBRIDGE;
				statusProperty.set("ONBRIDGE (" + timeWaited + ")");
				xPos += velocityX;
			}
			break;
		case ONBRIDGE:
			if (fromWest && xPos < bridgeEnd || !fromWest && bridgeEnd < xPos) {
				xPos += velocityX;
			} else {
				status = VStatus.PASSED;
				statusProperty.set("PASSED (" + timeWaited + ")");
				synchronized (this) {
					this.notify();
				}
				xPos += velocityX;
			}
			break;
		case PASSED:
			if (fromWest && xPos < landscapeEnd || !fromWest && landscapeEnd < xPos) {
				xPos += velocityX;
			} else {
				status = VStatus.DRIVENAWAY;
				closeInfoDialog();
			}
			break;
		case SHOCKED:
			break;
		case FALLING:
			xPos += velocityX;
			yPos += 0.08 * (accelerationY++);
			break;
		default:
			break;
		}

		if (corsaB) {
			if (1 <= jumpCounter && jumpCounter <= 7) {
				yPos -= velocityY;
			} else if (10 <= jumpCounter && jumpCounter <= 16) {
				yPos += velocityY;
			} else if (40 <= jumpCounter) {
				jumpCounter = 0;
			}
			jumpCounter++;
		}
		
		setX(xPos);
		setY(yPos);
		
		return status;
	}

	
	/**
	 * Requestor class which actually asks for the permission to cross the bridge.
	 * (So the refresh function isn't blocked). After the permission is granted, 
	 * the Requestor waits for a notify. Once the Requestor is notified (by the 
	 * refreh-function), the leaveBridge method is called. 
	 */
	class BridgeControlRequestor implements Runnable {

		@Override
		public void run() {
			bridgeControl.requestCrossing(VehicleImpl.this);
			paused = false;
			synchronized (VehicleImpl.this) {
				try {
					VehicleImpl.this.wait();
				} catch (InterruptedException e) {
					e.printStackTrace();
				}
			}
			bridgeControl.leaveBridge(VehicleImpl.this);
		}

	}

	/**
	 * Creates the little info dialog which shows up, when the user
	 * clicks on the vehicle. 
	 * 
	 * @return	Stage of the info dialog.
	 */
	Stage createInfoDialog() {

		Stage stage = new Stage();
		stage.setTitle("Vehicle #" + id);

		Label statusLB = new Label();
		statusLB.textProperty().bind(statusProperty);
		statusLB.setStyle("-fx-font: 18 verdana;");
		statusLB.setTextFill(Color.BLUE);
		BorderPane.setMargin(statusLB, new Insets(5, 5, 5, 5));

		TextArea messageTA = new TextArea();
		messageTA.textProperty().bind(messageProperty);
		messageTA.setEditable(false);
		BorderPane.setMargin(messageTA, new Insets(0, 5, 5, 5));

		BorderPane root = new BorderPane();
		root.setCenter(messageTA);
		root.setTop(statusLB);

		stage.setOnCloseRequest(new EventHandler<WindowEvent>() {
			public void handle(WindowEvent we) {
				infoDialogOpened = false;
			}
		});

		stage.setScene(new Scene(root, 360, 200));
		return stage;

	}

}
